#include "san.h"

#include <cstring>

#include "move.h"
#include "piece.h"
#include "square.h"
#include "move_legal.h"
#include "move_check.h"

const char pawn_char = 'P';
const char knight_char = 'N';
const char bishop_char = 'B';
const char rook_char = 'R';
const char queen_char = 'Q';
const char king_char = 'K';


/**
 * Checks if the current position is the second repetition
 * or if they were 50 moves without pawn move or capture
 */
bool isSecondRepetition(const board_t *board){
	if (board->ply_nb >= 100) { // potential draw
		if (board->ply_nb > 100) return true;
     	ASSERT(board->ply_nb==100);
      	return !board_is_mate(board);
   	}
		
	bool rep = false;
	for (int i = 4; i <= board->ply_nb; i += 2) {
		if (board->stack[board->sp-i] == board->key){
			if (rep == false) rep = true;
			else return true;
		}
	}
	return false;
}

/**
 * Generates and checks if the move from sqFrom to sqTo is legal. 
 * Returns the move or MoveNone if illegal
 */
int check_move(board_t * board, const int sqFrom64, const int sqTo64){
	int goodMove = MoveNone;
	undo_t undo[1];
	int m = MOVE_MAKE(SQUARE_FROM_64(sqFrom64), SQUARE_FROM_64(sqTo64));
	if (move_is_pseudo(m, board) && pseudo_is_legal(m, board)) goodMove = m;
	else{
		m = MOVE_MAKE_FLAGS(SQUARE_FROM_64(sqFrom64), SQUARE_FROM_64(sqTo64), MoveCastle);
		if (move_is_pseudo(m, board) && pseudo_is_legal(m, board)) goodMove = m;
		else{
			m = MOVE_MAKE_FLAGS(SQUARE_FROM_64(sqFrom64), SQUARE_FROM_64(sqTo64), MoveEnPassant);
			if (move_is_pseudo(m, board) && pseudo_is_legal(m, board)) goodMove = m;
			else{
				m = MOVE_MAKE_FLAGS(SQUARE_FROM_64(sqFrom64), SQUARE_FROM_64(sqTo64), MovePromoteQueen);
				if (move_is_pseudo(m, board) && pseudo_is_legal(m, board)) goodMove = m;
			}
		}
	}
	if (goodMove != MoveNone){
		move_do (board, m, undo);
		if (!board_is_legal(board)){
			goodMove = MoveNone;
		}
		move_undo(board, m, undo);
	}
	return goodMove;
}

void getStrfromPiece(char * str, const int piece) {
	str[1] = '\0';
	if (PIECE_IS_BISHOP(piece)) str[0] = bishop_char;
	else if (PIECE_IS_KING(piece)) str[0] = king_char;
	else if (PIECE_IS_KNIGHT(piece)) str[0] = knight_char;
	else if (PIECE_IS_PAWN(piece)) str[0] = pawn_char;
	else if (PIECE_IS_QUEEN(piece)) str[0] = queen_char;
	else if (PIECE_IS_ROOK(piece)) str[0] = rook_char;
}

/**
 * Writes the san string for the given move
 * on the given board
 * san: string where the san is written to, should be at least of size 8 
 * board: board before move is made
 */
void getSan(char * san, int move, board_t *board) {
	char p[2];
	char fromStr[3];	
	char toStr[3];
	
	const int from = MOVE_FROM(move);
	const int to = MOVE_TO(move);
	const int piece = MOVE_PIECE(move, board);
	square_to_string(from, fromStr, 3);
	square_to_string(to, toStr, 3);
	
	if (PIECE_IS_PAWN(piece)) {
		if (move_is_capture(move, board) || MOVE_IS_EN_PASSANT(move)) {
			san[0] = fromStr[0];
			san[1] = '\0';
			strcat(san, "x");
			strcat(san, toStr);
		}
		else {
			strcpy(san, toStr);
		}
		if (MOVE_IS_PROMOTE(move)) {
			strcat(san, "=");
			getStrfromPiece(p, move_promote(move));
			strcat(san, p);
		}
	}
	else if (MOVE_IS_CASTLE(move)) {
		if (to == square_from_string("c1") || to == square_from_string("c8"))
			strcpy(san, "O-O-O");
		else strcpy(san, "O-O");
	}
	else {
		int amb = 0;

		getStrfromPiece(p, piece);
		strcpy(san, p);
		
		//Check all squares for ambiguity
		for (int i=0; i<64; i++) {
			int sq = SQUARE_FROM_64(i);
			if (sq != from && board->square[sq] == piece) {
				int testmove = check_move(board, i, SQUARE_TO_64(to));
				if (testmove != MoveNone) {
					if (amb == 0) amb = 1;
					if (SQUARE_FILE(sq) == SQUARE_FILE(from)) { //a-h
						if (amb == 2) amb = 4;
						else amb = 3;
					}
					else if (SQUARE_RANK(sq) == SQUARE_RANK(from)) { //1-8
						if (amb == 3) amb = 4;
						else amb = 2;
					}
				}
			}
		}
		if (amb == 1 || amb == 2) strncat(san, fromStr, 1);
		else if (amb == 3) strncat(san, &(fromStr[1]), 1);
		else if (amb == 4) strcat(san, fromStr);
		
		
		if (move_is_capture(move, board)) strcat(san, "x");
		
		strcat(san, toStr);
	}
	undo_t undo[1];
	move_do(board, move, undo);
	if (board_is_mate(board)) strcat(san, "#");
	else if (board_is_check(board)) strcat(san, "+");
	move_undo(board, move, undo);
}

void getSan(char * san, board_t *board, const int *moves, const undo_t *undos, int nr) {
	const int l = board->sp;
	if (nr < 0) {
		for (int i=0; i<-nr; i++) {
			move_undo(board, moves[l-1-i], &undos[l-1-i]);
		}
		getSan(san, moves[l + nr], board);
		undo_t u[1];
		for (int i=-nr; i>0; i--) {
			move_do(board, moves[l-i], u);
		}
		return;
	}
	else {
		for (int i=0; i<l-nr; i++) {
			move_undo(board, moves[l-1-i], &undos[l-1-i]);
		}
		getSan(san, moves[nr], board);
		undo_t u[1];
		for (int i=nr; i<l; i++) {
			move_do(board, moves[i], u);
		}
		return;
	}	
}